/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                                       *
 *       Filename:  leslieMat.c                                                          *
 *                                                                                       *
 *    Description:  Using a matrix representation of a target population, the            *
 *                    program calculates the projected population after a number         *
 *                    of generations, as well as the eigenvalue corresponding to         *
 *                    population growth rate.                                            *
 *                                                                                       *
 * Specifications:  Uses double precision values for precision, and malloc               *
 *                    for memory allocation requests. Also, the matrix                   *
 *                    representation is 1-dimensional and uses pointer                   *
 *                    arithmatic for access to elements.                                 *
 *                                                                                       *
 *            Due:  06/14/2010                                                           *
 *                                                                                       *
 *        Authors:  Jesse A. Hanley                                                      *
 *                                                                                       *
 *   Contact Info:  HanleyJA@email.wofford.edu                                           *
 *                                                                                       *
 *   Command Line                                                                        *
 *     Arguments:   The following are some of the command line arguments the             *
 *                    program accepts ( -h is not currently supported )                  *
 *                                                                                       *
 *                  -o <filename> :   Supply a file for output to be displayed.          *
 *                  -i <filename> :   Supply a file for input matrix to be read.         *
 *                  -g <#>        :   Supply the number of generations to be ran.        *
 *                  -d <filename> :   Supply a file for the age distribution.            *
 *                  -h            :   Get help about this program, including             *
 *                                     input format.                                     *
 *                                                                                       *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <math.h>
/************************* 1 *************************/
#include "leslieMat.h"

int main(int argc, char **argv)
{
   int iSize, iWidth, iRank, iProcSize, aiTemp[3], iGenerations, statusContinue = 1;

   double *ptrValueArray, *ptrDistArray, *ptrMatrixPower, *ptrResultValues, dEigenValue = 0.0;
   double *ptrNewDist1, *ptrNewDist2, *ptrNewDist3, dChange = 0.0, dOldValue = 0.0;
   char *sFileInput, *sFileDist, *sFileOutput;

   /************************* 2 *************************/
   
   

   /*
    *  Check the command line arguments, and exit if not supplied correctly.
    */
   /************************* 3 *************************/
   
   parse_arguments(argc, argv, &sFileInput, &sFileDist, &sFileOutput, &iGenerations);
   if (( iGenerations == 0   ) ||
       ( sFileInput == NULL  ) ||
       ( sFileDist == NULL   ) ||
       ( sFileOutput == NULL ) )
   {
      statusContinue = 0;
   }
   /*
      fprintf( stderr, " iGenerations: %d\n sFileInput: %s\n sFileDist: %s\n sFileOutput: %s\n\n",
      iGenerations, sFileInput, sFileDist, sFileOutput );
    */
   /************************* 3 *************************/

   /************************* 4 *************************/

   if ( statusContinue == 0 )
   {
      /************************* 5 *************************/
      
      print_usage();
      

      /************************* 5 *************************/
      exit(1);
   }

   /*
    *  Load the values from the input file(s) into the appropriate arrays
    */
   /************************* 6 *************************/
   
   load_values(&ptrValueArray, &iSize, &iWidth, sFileInput);
   
   
   
   /************************* 6 *************************/

   /*
    *  Broadcasts these values as a list.  It is more efficient to package
    *    them in a list and broadcast once, than to perform two collective
    *    communications.
    */
   /************************* 7 *************************/
   
   
   
   
   
   
   
   /************************* 7 *************************/

   malloc_double_list( &ptrMatrixPower, iSize );

   /************************* 8 *************************/
   
   query_distribution( &ptrDistArray, iWidth, sFileDist );
   
   
   
   
   
   /************************* 8 *************************/

   /*
    *  Perform the actual work, raising the matrix to a power,
    *    and then multiplying it by the given distribution
    */
   array_to_power(&ptrValueArray, &ptrMatrixPower, iWidth, iGenerations, iRank, iProcSize );

   /************************* 11 *************************/
   
   mat_product_vector( &ptrMatrixPower, &ptrDistArray, &ptrResultValues, iWidth );
   
   
   
   
   
   /************************* 11 *************************/
   
   /*
    *  Continue further, finding the eigenvalue for the system.
    */
   /************************* 12 *************************/

   malloc_double_list( &ptrNewDist1, iWidth );
   malloc_double_list( &ptrNewDist2, iWidth );

   list_copy( &ptrResultValues, &ptrNewDist1, iWidth );

   mat_product_vector( &ptrValueArray, &ptrNewDist1, &ptrNewDist2, iWidth );
   mat_product_vector( &ptrValueArray, &ptrNewDist2, &ptrNewDist3, iWidth );

   dChange = (ptrNewDist3[0] / ptrNewDist2[0]) - (ptrNewDist2[0] / ptrNewDist1[0]);
   if ( dChange < 0 )
   {
      dChange *= -1;
   }

   while ( dChange > 0.02 )
   {
      dOldValue = ptrNewDist1[0];
      list_copy( &ptrNewDist2, &ptrNewDist1, iWidth );
      mat_product_vector( &ptrValueArray, &ptrNewDist1, &ptrNewDist2, iWidth );
      dChange = (ptrNewDist2[0] / ptrNewDist1[0]) - (ptrNewDist1[0] / dOldValue );
      if ( dChange < 0 )
      {
         dChange *= -1;
      }
   }

   dEigenValue = (ptrNewDist2[0] / ptrNewDist1[0]);

   /*
    *  Call the print function, which prints the results into 
    *    the file specified by the command line argument -o,
    *    but only use process 0.
    */

   /************************* 13 *************************/
   
   char sTitle[] = "Results: \n";

   print_results(sTitle, &ptrResultValues, iWidth, iGenerations, sFileInput, sFileDist, sFileOutput, dEigenValue );
   /************************* 13 *************************/

   free_double_list( &ptrNewDist1 );
   free_double_list( &ptrNewDist2 );
   free_double_list( &ptrMatrixPower );
   /************************* 14 *************************/

   return 0;
}

void print_usage()
{
   fprintf( stderr,
"Command Line Arguments: \n\
      -o <filename> :   Supply a file for output to be displayed.   \n\
      -i <filename> :   Supply a file for input matrix to be read.  \n\
      -g <#>        :   Supply the number of generations to be ran. \n\
      -d <filename> :   Supply a file for the age distribution.     \n\
      -h            :   Get help about this program, including      \n\
                                    input format.                             \n\
");
   exit(1);
}

/*
   Pre-Condition:
      ptrList       :   Pointer to empty 1-Dimenionsal Array
      iEntries      :   Number of entries to allocate in the array

   Post-Condition:
      ptrList is now an array <iEntries> long of doubles

   Purpose          :   This function allocates a list in memory
*/
void malloc_double_list(double **ptrList, int iEntries)
{
   *ptrList = malloc(sizeof(double) * iEntries); 
}

/*
   Pre-Condition:
      ptrList       :   Pointer to empty 1-Dimenionsal Array
      iEntries      :   Number of entries to allocate in the array

   Post-Condition:
      ptrList is now an array <iEntries> long of ints

   Purpose          :   This function allocates a list in memory
*/
void malloc_int_list( int **ptrList, int iEntries)
{
   *ptrList = malloc(sizeof(int) * iEntries);
}

/*
   Pre-Condition:
      ptrList       :   Pointer to 1-Dimenionsal Array, allocated
                                    with malloc (mallocList)

   Post-Condition:
      Memory pointed to by ptrList has been freed.

   Purpose          :   This function frees allocated memory.
*/
void free_double_list(double **ptrList)
{
   free(*ptrList);
}

/*
   Pre-Condition:
      ptrList       :   Pointer to 1-Dimenionsal Array, allocated
                                    with malloc (mallocList)

   Post-Condition:
      Memory pointed to by ptrList has been freed.

   Purpose          :   This function frees allocated memory.
*/
void free_int_list(int **ptrList)
{
   free(*ptrList);
}

/*
   Pre-Condition:
      ptrList       :   Pointer to 1-Dimenionsal Array, allocated
                                    with malloc (mallocList)

   Post-Condition:
      iItem is appended to the end of ptrList, and iSize is updated

   Purpose          :   This function allocates a new array one item
                        longer, copies the values over, and copies
                        in the new value to the last position.
*/
void append( int **ptrList, int *iSize, int iItem)
{
   int *newLstPtr, i;
   *iSize = (*iSize) + 1;

   malloc_int_list( &newLstPtr, (*iSize) );

   for( i = 0; i < (*iSize); i++)
   {
      newLstPtr[i] = (*ptrList)[i];
   }

   newLstPtr[(*iSize)] = iItem;
   free_int_list(ptrList);

   (*ptrList) = newLstPtr;

}

/*
   Pre-Condition:
      ptrFile       :   Pointer to file

   Post-Condition:
      The number of doubles in the file is returned, and the
         file pointer is rewound.

   Purpose          :   This function counts the number of
                        items in a file.
*/
int count_entries(FILE *ptrFile)
{
   int iCount = 0;
   double dValue;

   while ( feof(ptrFile) == 0 )
   {
      fscanf(ptrFile, "%lf", &dValue);
      iCount++;
   }
   rewind( ptrFile );
   return iCount;
}

/*
   Pre-Condition:
      ptrValueArray :   Pointer to array, already allocated
      ptrInputArray :   Pointer to array, already allocated

   Post-Condition:
      ptrInputArray now points to an array filled with the corresponding
         leslie matrix created by the values within the input file

   Purpose          :   This function makes ptrInputArray point to a new
                        allocated array, that is filled with appropriate
                        values of the corresponding leslie matrix.
*/
void fill_value_array( int iEntries, double **ptrValueArray, double **ptrInputArray )
{
   int i, j, iWidth, index = 0, valueIndex;

   iWidth = (iEntries + 1) / 2;

   for( i = 0; i <= iWidth; i++ )
   {
      (*ptrInputArray)[i] = (*ptrValueArray)[i];
   }
   
   index = i;
   valueIndex = i;

   for( i = 0; i < (iWidth - 2); i++ )
   {
      for( j = 0; j < iWidth; j++ )
      {
         (*ptrInputArray)[index + j] = 0;
      }
      index += j;
      (*ptrInputArray)[index] = (*ptrValueArray)[valueIndex];
      valueIndex++;
   }
   (*ptrInputArray)[index + 1] = 0;
}

/*
   Pre-Condition:
      ptrFile       :   Pointer to the input file
      ptrValueArray :   Pointer to alloced array

   Post-Condition:
      ptrValueArray now holds the different values for population information

   Purpose          :   This function reads in the values from a file and fills
                        ptrValueArray with these values.
*/
void read_values(FILE *ptrFile, int iEntries, double **ptrValueArray)
{
   double dTemp = 0;
   int i;

   for ( i = 0; i < iEntries; i++ )
   {	
      fscanf(ptrFile, "%lf", &dTemp);
      (*ptrValueArray)[i] = dTemp;
   }
}

/*  
   Pre-Condition: 
      ptrInputArray :   Pointer to an empty 1-Dimensional Array

   Post-Condition:
      ptrInputArray is loaded with values from the ptrFile

   Purpose          :   This function reads in the values from a file and fills
                     ptrInputArray with these values.
*/
void load_values(double **ptrInputArray, int *ptrSize, int *ptrWidth, char *sFileName)
{
   int iEntries, iSize;
   double *ptrValueArray;

   FILE *ptrFile;

   if ( (ptrFile = fopen(sFileName, "r")) == NULL )
   {
      fprintf( stderr, 
         "Error: Unable to open matrix value input file.\n" );
      print_usage();
   }

   iEntries = count_entries(ptrFile);
   (*ptrWidth) = ( (iEntries + 1) / 2 );
   iSize = (*ptrWidth) * (*ptrWidth);
   (*ptrSize) = iSize;
   
   malloc_double_list( &ptrValueArray, iEntries);
   malloc_double_list( ptrInputArray, iSize );

   read_values(ptrFile, iEntries, &ptrValueArray);

   fill_value_array( iEntries, &ptrValueArray, ptrInputArray );
   fclose(ptrFile);

}

/*  
   This is the command line argument parser. When supplied on runtime,
      they provide information crucial to the program, and allow the
      program to change without the need to recompile.
   -d <file> :   The distribution file, with the initial population.
   -g <#>    :   The number of generations to run.
   -i <file> :   The input file, regarding survival probability,
                           and the reproduction rate.
   -o <file> :   The output file, where the results are printed.
   -h        :   Not implemented yet.  Will be once documentation is
                           completed.
*/
void parse_arguments(int argc, char **argv, char **sFileInput, char **sFileDist, char **sFileOutput, int *iGenerations )
{
   int iVal;
   while ((iVal = getopt(argc, argv, "d:g:i:o:")) != -1) 
   {
      switch (iVal)
      {
      case 'd':
         (*sFileDist) = malloc(sizeof(optarg) );
         strcpy( (*sFileDist), optarg );
         break;
      case 'g':
         (*iGenerations) = atoi(optarg);
         break;
      case 'i':
         (*sFileInput) = malloc(sizeof(optarg) );
         strcpy( (*sFileInput), optarg );
         break;
      case 'o':
         (*sFileOutput) = malloc(sizeof(optarg) );
         strcpy( (*sFileOutput), optarg );
         break;
      default:
         break;
      }
   }
}

/*  
   Pre-Condition: 
      ptrInputArray :   Pointer to an empty 1-Dimensional Array

   Post-Condition:
      ptrInputArray is loaded with values from the distribution file

   Purpose:
      This function fills ptrInputArray with the values from sFileDist
*/
void query_distribution( double **ptrInputArray, int iArrayWidth, char *sFileDist )
{
   int i;
   double dTemp = 0;
   FILE *ptrFile;

   if ( (ptrFile = fopen( (sFileDist), "r")) == NULL )
   {
      fprintf( stderr, 
         "Error: Unable to open distribution file.\n" );
      print_usage();
   }
   malloc_double_list( ptrInputArray, iArrayWidth );

   for ( i = 0; i < iArrayWidth; i++ )
   {
      fscanf(ptrFile, "%lf", &dTemp);
      (*ptrInputArray)[i] = dTemp;
   }

}

/*  
   Pre-Condition: 
      ptrArray      :   Pointer to an 1-Dimensional Array
      ptrVector     :   Pointer to a vector of values

   Post-Condition:
      ptrResults now contains the result from the multiplication of the inputs

   Purpose:
      This function multiplies ptrArray by ptrVector and places the results in ptrResults
*/
void mat_product_vector( double **ptrArray, double **ptrVector, double **ptrResults, int iWidth )
{
   int i, j;
   int iResultSize = sizeof((*ptrResults)) / sizeof(double);
   int iVectorSize = sizeof((*ptrVector))  / sizeof(double);

   malloc_double_list( ptrResults, iVectorSize );

   for ( i = 0; i < iResultSize; i++ )
   {
      (*ptrResults)[i] = 0;
   }

   for ( i = 0; i < iWidth; i++ )
   {
         for ( j = 0; j < iWidth; j++ )
         {
               (*ptrResults)[i] += ( (*ptrArray)[ i*iWidth + j ] ) * ( (*ptrVector)[j] );
         }
   }
}

/*  
   Pre-Condition: 
      ptrArray1      :   Pointer to an 1-Dimensional Array
      ptrArray2      :   Pointer to an 1-Dimensional Array

   Post-Condition:
      ptrResults now contains the result from the multiplication of the inputs

   Purpose:
      This function multiplies ptrArray1 by ptrArray2 and places the results in ptrResults
*/
void mat_product_mat( double **ptrArray1, double **ptrArray2, double **ptrResults, int iWidth, int iRank, int iSize )
{
   int i, j, k;

   for ( i = 0; i < iWidth*iWidth; i++ )
   {
      (*ptrResults)[i] = 0;
   }
   /************************* 9 *************************/
   for ( i = 0; i < iWidth; i++ )
   {
         for ( j = 0; j < iWidth; j++ )
         {
               for ( k = 0; k < iWidth; k++ )
               {
                  (*ptrResults)[ i*iWidth + j] += ((*ptrArray1)[ i*iWidth + k ]) * ((*ptrArray2)[ k*iWidth + j ]);
               }
         }
   }

   /************************* 10 *************************/
}

/*  
   Pre-Condition: 
      inputArray      :   Pointer to an 1-Dimensional Array

   Post-Condition:
      ptrResultArray now contains the result from raising the array to a power

   Purpose:
      This function does the multiplication to raise inputArray to the iPower
*/
void array_to_power(double **inputArray, double **ptrResultArray, int iWidth, int iPower, int iRank, int iSize)
{
   int i, j;
   double *ptrTempArray1, *ptrTempArray2;

   /* To the First power */

   if ( iPower == 1 )
   {
      for( i = 0; i < ( iWidth*iWidth ); i++ )
      {
         (*ptrResultArray)[i] = (*inputArray)[i];
      }
   }
   else
   {
      malloc_double_list( &ptrTempArray1, ( iWidth * iWidth ) );
      malloc_double_list( &ptrTempArray2, ( iWidth * iWidth ) );

      for( i = 0; i < ( iWidth * iWidth ); i++ )
      {
         ptrTempArray1[i] = (*inputArray)[i];
         ptrTempArray2[i] = 0;
      }

      for( i = 1; i < iPower; i++ )
      {
         mat_product_mat( inputArray, &ptrTempArray1, &ptrTempArray2, iWidth, iRank, iSize );
         for( j = 0; j < ( iWidth * iWidth ); j++ )
         {
            ptrTempArray1[j] = ptrTempArray2[j];
         }
      }
      for( i = 0; i < ( iWidth * iWidth ); i++ )
      {
         (*ptrResultArray)[i] = ptrTempArray1[i];
      }
   }
}

/*
   Pre-Condition:
      The program has ran, and the various inputs have proper values

   Post-Condition:
      The results of the run have been printed to sOutputFile

   Purpose:
      This function prints the results of the run to an output file.
*/
void print_results( char *ptrCharTitle, double **ptrValues, int iWidth, int iGenerations, 
                   char *sInputFile, char *sDistFile, char *sOutputFile, double dEigenValue )
{
   int i;
   FILE *ptrFile;
   double dTotal = 0.0;

   if ( (ptrFile = fopen( sOutputFile, "w")) == NULL )
   {
      fprintf( stderr, 
         "Error: Unable to open output file for writing.\n" );
      print_usage();
   }

   fprintf( ptrFile, ptrCharTitle );
   fprintf( ptrFile, "    Generations:       %d \n", iGenerations       );
   fprintf( ptrFile, "    Input File:        %s \n", sInputFile         );
   fprintf( ptrFile, "    Distribution File: %s \n", sDistFile          );
   fprintf( ptrFile, "    Output File:       %s \n", sOutputFile        );
   fprintf( ptrFile, "\nFinal Values:         \n\n" );

   for ( i = 0; i < iWidth; i++ )
   {
      fprintf( ptrFile, "    %lf\n", ( (*ptrValues)[i] ) );
      dTotal += (*ptrValues)[i];
   }
   fprintf( ptrFile, "\nFinal %% Distribution:\n\n" );
   for ( i = 0; i < iWidth; i++ )
   {
      fprintf( ptrFile, "    %lf %%\n", (( (*ptrValues)[i] )/ dTotal) );
   }
   fprintf( ptrFile, "\nEigenvalue:        %lf\n",  dEigenValue     );
   fprintf( ptrFile, "Population Growth: %5.2lf\n", dEigenValue*100 );
   fclose( ptrFile );
}

/*  
   Pre-Condition: 
      ptrList1      :   Pointer to an 1-Dimensional Array
      ptrList2      :   Pointer to an 1-Dimensional Array

   Post-Condition:
      The values of ptrList1 have been copied into ptrList2.

   Purpose:  This function copies the values of one array into another
*/
void list_copy( double **ptrList1, double **ptrList2, int iWidth )
{
   int i;
   for( i = 0; i < iWidth; i++ )
   {
      (*ptrList2)[i] = (*ptrList1)[i];
   }
}
